'
' Goattracker replay routine - early alfa test version
'

CON _clkmode = xtal1 + pll16x
    _xinfreq = 5_000_000    
 
VAR

  byte tunes
  byte instruments
  byte patterns

  byte waveTablePos[3]   
  byte pulseTablePos[3]
  byte filterTablePos[3] 
  byte speedTablePos[3] 

  byte waveTableLength
  byte pulseTableLength
  byte filterTableLength
  byte speedTableLength
  byte orderLength[3]

  byte orderOffset[3]  
  byte note[3]
  long PWM_Add[3]
  byte PWM_Timer[3]
  word PWM_Value[3]

  byte cutoffTimer 
  long cutoffAdd
  long cutoffValue

  long patternDataPointer
  long instrumentPointer
  long waveTablePointer 
  long pulseTablePointer 
  long filterTablePointer 
  long speedTablePointer 

  long transpose[3]
  long notePointer[3]
  long orderPointer[3]
  
  byte patternTickPeriod
  byte patternTickCounter  
  long frameCounter

  long SID_Pointer
  long song

  byte playing
  byte playFx[3]
    
OBJ
  SID : "SIDcog_Lite"  

PUB start( right, left )
  SID_Pointer := SID.start( left, right )
  SID.setVolume( 15 )
  patternTickCounter := patternTickPeriod := 5
  playing := false
   
PUB setSong( songPointer )
  song := songPointer

PUB stopSong | i
  playing := false
  repeat i from 0 to 23
    SID.setRegister( i, 0 )     

PUB setTempo( tempo )
  patternTickCounter := patternTickPeriod := tempo   

PUB playSoundFx( FxNumber )
  playFx[2] := FxNumber 
  
PUB tick | i, orderData, gotNewPattern, freq, noteValue, instrument, command, commandData, temp, signedByte  

  ifnot playing
    return 

  if patternTickCounter-- == 0
    patternTickCounter := patternTickPeriod

  if patternTickCounter == patternTickPeriod or playFx[2]  
    repeat i from 0 to 2
      notePointer[i] += 4
      
      ' Look for end of pattern. If it is reached, do the right thing!
      if byte[ song + notePointer[i] ] == $ff
        ' Iterate until we have got a new pattern or reached the end of the song
        gotNewPattern := false
        repeat until gotNewPattern 
          orderData := byte[ song + orderPointer[i] + orderOffset[i]++ ]

          if orderData < $D0                               ' Is it a pattern change command ?    
            gotNewPattern := true 
          elseif orderData < $E0                           ' Else, is it a repeat command ? 
          elseif orderData < $ff                           ' Else, is it a transpose command ?
            transpose[i] := orderData - $F0
          else                                             ' Else, it is the end of the song  
            orderOffset[i] := byte[ song + orderPointer[i] + orderOffset[i] ] 
        ' Set the new notePointer to the first step of the new pattern
        notePointer[i] := getPatternPointer( orderData ) + 1
 
'     Update row data       
      temp       := song + notePointer[i] 
      noteValue  := byte[temp++]
      instrument := byte[temp++]
      'command    := byte[temp++]
      'commandData:= byte[temp++]

      ' Handle sound FX
      if playFx[i]
        instrument := playFx[i]
        playFx[i] := false
        
      ' Handle instrument settings         
      ifnot instrument == 0
        instrument--
        temp := song + instrumentPointer + (instrument*25)
        byte[SID_Pointer+(i<<3)+5] := byte[temp++] ' Set Attack/Decay 
        byte[SID_Pointer+(i<<3)+6] := byte[temp++] ' Set Sustain/Release  
        waveTablePos[i]   := byte[temp++]  
        pulseTablePos[i]  := byte[temp++]
        filterTablePos[i] := byte[temp++]
        speedTablePos[i]  := byte[temp++]

        if noteValue == $bd                      ' Is it a rest command ?      
        elseif noteValue == $be                  ' Else, is it a keyoff command ?  
          SID.noteOff(i)                  
        elseif noteValue == $be                  ' Else, is it a keyon command ?   
        elseif noteValue>$5f and noteValue<$bd   ' Else, it is a note command !
          SID.noteOff(i) 
          note[i] := noteValue-$60
      else   
        if noteValue == $be
          SID.noteOff(i) 
        if noteValue>$5f and noteValue<$bd
          note[i] := noteValue-$60
          
' Handle frame updates 
'----------------------------------------------------------------------------------------------  
  frameCounter++ 
  repeat i from 0 to 2

  ' Handle wave-table values and increment position 
'----------------------------------------------------------------------------------------------   
    if waveTablePos[i] > 0
      if (temp := byte[ song + waveTablePointer + waveTablePos[i] ]) == $ff
        waveTablePos[i] := byte[ song + waveTablePointer + waveTablePos[i] + waveTableLength ] 
      else
        SID.setRegister( (i*7) + 4, temp )

        ' Handle pitch    
        signedByte := byte[ song + waveTablePointer + waveTablePos[i] + waveTableLength ] 
        if signedByte > $80
          freq := note2Freq( (signedByte-$81) ) 
          SID.setFreq(i, freq ) 
        elseif signedByte < $80
          if signedByte&$40
            signedByte -= $80
          freq := note2Freq( note[i] + transpose[i] + signedByte )
          SID.setFreq(i, freq ) 
        waveTablePos[i]++

  ' Handle pulse-table values and increment position  
'---------------------------------------------------------------------------------------------- 
    if pulseTablePos[i] > 0
      if (temp := byte[ song + pulseTablePointer + pulseTablePos[i] ]) == $ff
        pulseTablePos[i] := byte[ song + pulseTablePointer + pulseTablePos[i] + pulseTableLength ] 
      else
      ' Handle absolute PWM values
        if temp&$80
          PWM_Value[i] := ((temp&$f)<<8) | byte[ song + pulseTablePointer + pulseTablePos[i] + pulseTableLength ]
          pulseTablePos[i]++ 

        'Handle relative PWM add-values
        else
          if PWM_Timer[i] == 0
            PWM_Timer[i] := temp
            signedByte := byte[ song + pulseTablePointer + pulseTablePos[i] + pulseTableLength ]   
            ' Handle signed byte value
            if signedByte&$80
              signedByte -= $100
            PWM_Add[i] := signedByte 
            pulseTablePos[i]++
            
          PWM_Timer[i]--
          PWM_Value[i] += PWM_Add[i]        

      SID.setPWM( i, PWM_Value[i] )
      
  ' Handle filter-table values and increment position  
'---------------------------------------------------------------------------------------------- 
    if filterTablePos[i] > 0
      if (temp := byte[ song + filterTablePointer + filterTablePos[i] ]) == $ff
        filterTablePos[i] := byte[ song + filterTablePointer + filterTablePos[i] + filterTableLength ] 

      else
        ' Handle filter settings
        if temp&$80
          SID.setFilterType(temp,temp,temp)
          temp := byte[ song + filterTablePointer + filterTablePos[i] + filterTableLength ]
          SID.setResonance( temp >> 4 ) 
          SID.setFilterMask( temp, temp, temp ) 
          filterTablePos[i]++
          
        ' Handle absolute cutoff frequency
        elseif temp == 0
          cutoffValue := byte[ song + filterTablePointer + filterTablePos[i] + filterTableLength ] << 3 
          SID.setCutoff( cutoffValue )
          filterTablePos[i]++
          
        ' Handle relative cutoff add-values
        else
          if cutoffTimer == 0
            cutoffTimer := temp
            signedByte := byte[ song + filterTablePointer + filterTablePos[i] + filterTableLength ]   
            ' Handle signed byte value
            if signedByte&$80
              signedByte -= $100
              
            cutoffAdd := signedByte 
            filterTablePos[i]++
            
          cutoffTimer--
          cutoffValue := (cutoffValue + cutoffAdd ) #> 0
          cutoffValue := cutoffValue <# $7ff
          SID.setCutoff( cutoffValue )

  '!outa[16]
  return
 
PUB initSong | pointer, tempValue, i, gotNewPattern, orderData
  ' Check if the "file" is valid
  ifnot byte[ song ] == "G" and byte[ song + 1] == "T" and byte[ song +2] == "S" and byte[ song +3] == "5"
    return -1 ' This isn't a valid Goat Tracker module
  if (tunes := byte[ song + 100])>1
    return -2 ' Multiple subtunes are not supported

'                 Read "file" and get some pointers     
'-------------------------------------------------------------------- 
  pointer := 101
  repeat i from 0 to 2
    orderPointer[i] := pointer+1 
    tempValue := byte[ song + pointer]
    pointer += tempValue + 2     
    orderLength[i] := tempValue  

  instruments := byte[ song + pointer]
  pointer++
  instrumentPointer := pointer
  pointer += 25*instruments
  waveTablePointer := pointer 
  waveTableLength  := byte[ song + pointer]
  pointer += (waveTableLength<<1)+1 
  pulseTablePointer := pointer 
  pulseTableLength  := byte[ song + pointer] 
  pointer += (pulseTableLength<<1)+1 
  filterTablePointer := pointer 
  filterTableLength  := byte[ song + pointer] 
  pointer += (filterTableLength<<1)+1 
  speedTablePointer := pointer  
  speedTableLength  := byte[ song + pointer] 
  pointer += (speedTableLength<<1)+1 
  patternDataPointer := pointer+1
  patterns := byte[ song + pointer]

'                 Init note pointers and other stuff
'-------------------------------------------------------------------- 
  repeat i from 0 to 2
    transpose[i]   := 0
    PWM_Timer[i]   := 0
    PWM_Add[i]     := 0
    cutoffTimer    := 0
    orderOffset[i] := 0
    
    gotNewPattern := false
    ' Iterate until we have got a new pattern and not any transpose commands and such  
    repeat until gotNewPattern 
      orderData := byte[ song + orderPointer[i] + orderOffset[i]++ ]
      if orderData < $D0                   ' Is it a pattern a new pattern command ? 
        gotNewPattern := true 
      elseif orderData < $E0               ' Else, is it a repeat command ?  
      elseif orderData < $ff               ' Else, is it a transpose command ?
        transpose[i] := orderData - $F0
      else                                 ' Else, it is the end of the song

    notePointer[i] := getPatternPointer( orderData ) + 1     
 
  playing := true     
  
  return 0
 
PUB getPatternPointer(patternNumber) | tempValue, i

  if patternNumber<1
    return patternDataPointer

  ' Get the right pattern pointer by successively read and add to the pointer "patternNumber of times"  
  tempValue := patternDataPointer
  patternNumber--
  repeat i from 0 to patternNumber
    tempValue += (byte[ song + tempValue]<<2)+1

  return tempValue 
 
PUB note2freq(noteValue) | octave
    noteValue += 24
    octave := (noteValue/12)
    noteValue -= octave*12 
    return (noteTable[noteValue]>>(8-octave))

DAT noteTable word 16350, 17320, 18350, 19450, 20600, 21830, 23120, 24500, 25960, 27500, 29140, 30870
 